
########################################################################################
###         This file contains R codes to 
###              perform Multilevel FPCA (Functional Principal Component Analysis)
###                                     
###             Author:     Chongzhi Di
###             Based on the work with Ciprian M.  Crainiceanu 
###                                                             
###             The corresponding paper is under review
###             Date:       07/16/2007
########################################################################################


########################################################################################
### 
### Description:
###     The function "MFPCA" is designed to perform multilevel functional principal 
###     component analysis. The methodology aim to separate the variation in multilevel
###     functional data into two parts, between cluster and within cluster variations. 
###     The principal directions of variation at each level is characterized by a set of 
###     eigenfunctions, ordered by the eigenvalues that indicates the magnate of variation.
###     The method is developed and explained in details in the ongoing paper "Multilevel 
###     Functional Data Analysis". 
###
###     The function can deal with functions that are measured with or without noise by 
###     specifying the "smoothing" parameter. An alternate way to deal with measurement 
###     error is to smooth each individual functions first before carrying out the multilevel
###     functional principal component analysis. 
###
###     Currently, the function only deals with functional data that is measured on regular 
###     grid points, i.e, every function is measured at the same set of grid points. But it 
###     should be easy to modify the codes to deal with irregular grid points. 
###
########################################################################################


########################################################################################
###
### Input of the function "MFPCA"
###
### y:      a data matrix containing functional responses. Each row contains measurements 
###         from a function at a set of grid points, and each column contains measurements
###         of all functions at a particular grid point.
### id:     a vector contains the id information that is used to identify clusters. 
### visit:  optional argument if functions from the same cluster have natural orders 
###         such as visit time.
### M:      number of clusters  
### J:      cluster size
### N:      number of grid points per function. 
### tstart: the start point of the grid points
### tlength: the length of the interval that the functions are evaluated at.
### twoway: a logical argument indicates whether oneway or twoway functional ANOVA (analysis
###         of variance) decomposition is more proper for the problem. "twoway=TRUE" will carry
###         out twoway ANOVA, and the default, "twoway=TRUE" will carry out twoway ANOVA.
### smoothing: a logical argument that controls whether or not smoothing is conducted. When 
###         the functional data is measured without noise, smoothing is not needed. On the other
###         hand, if measurement error is present, specifying "smoothing=TRUE" will incorporate 
###         smoothing steps. As a result, the estimated eigenvalues will be less biased and the
###         estimated eigenfunctions will be more smooth. The default is "smoothing=FALSE".
### min.K1: the minimum number of components to keep in level 1 space. The actual number of 
###         components will be the maximum of min.K1 and the number determined by the data.
### min.K2: the minimum number of components to keep in level 2 space.
### 
########################################################################################


########################################################################################
###
### Output of the function "MFPCA"
### 
###     The output of the function is a list that contains the following elements. 
###
### K1:         the number of components at level 1
### K2:         the number of components at level 2
### lambda1:    a vector containing all level 1 eigenvalues in non-increasing order
### lambda2:    a vector containing all level 2 eigenvalues in non-increasing order
### phi1:       a matrix containing all level 1 eigenfunctions. Each row contains an 
###             eigenfunction evaluated at the same set of grid points as the input data.
###             The eigenfunctions are in the same order as the corresponding eigenvalues.
### phi2:       a matrix containing all level 2 eigenfunctions. Each row contains an 
###             eigenfunction evaluated at the same set of grid points as the input data.
###             The eigenfunctions are in the same order as the corresponding eigenvalues.
### scores1:    a matrix containing estimated level 1 principal component scores. Each row
###             corresponding to the level 1 scores for a particular subject in a cluster.
###             The number of rows is the same as that of the input matrix y. Each column 
###             contains the scores for a level 1 component for all subjects.
### scores2:    a matrix containing estimated level 2 principal component scores. Each row
###             corresponding to the level 2 scores for a particular subject in a cluster.
###             The number of rows is the same as that of the input matrix y. Each column 
###             contains the scores for a level 2 component for all subjects.
### mu:         a vector containing the overall mean function.
### eta:        a matrix containing the deviation from overall mean function to visit specific 
###             mean function. The number of rows is the number of visits.
###
#########################################################################################


MFPCA <- function(y, id=NULL, visit=NULL, M=NULL, J=NULL, N=NULL, tstart=0, tlength=1, twoway=TRUE,
                            smoothing=FALSE, min.K1=4, min.K2=4) {

###     check the missingness of arguments
###     1. Either (id,visit) or (M,J) is needed for identifying clusters. 
###         If both are missing, give error message "not enough information";
###         if only (M,J) is provided, the function would generate (id,visit);
###         if only ((id,visit) is provided, the function would generate (M,J);
###      Note:  currently the function only deals with balanced data, i.e, each 
###             subject is measured at the same J visits.
###     2. If the number of grid points N is missing, the function would generate it.

    if(  ( is.null(id)|is.null(visit) ) && ( is.null(M)|is.null(J) )  )
        stop("Not enough information! Please provide (id, visit) or (M,J) !")
    if( !( is.null(M)|is.null(J) ) && ( is.null(id) | is.null(visit) ) ) {
        id <- rep(1:M,each=J)
        visit <- rep(1:J, M)
    }
    if( !( is.null(id) | is.null(visit) ) && (is.null(M) | is.null(J) ) )  {
        M <- length(table(id))
        J <- length(table(visit))        
    }
    if(is.null(N)) 
        N <- dim(y)[2]

###     Generate the set of grid points that are equally spaced.
    t <- seq(tstart, tstart + tlength, length=N)
    
###     Calculate overall mean function. 
###     If twoway functional ANOVA is needed ("twoway=TRUE"),  the visit specific mean function
###     and the deviation from the overall mean to visit specific mean functions are also
###     computed. 
    mu <- apply(y, 2,mean)
    eta <- matrix(0, J, N)
    if(twoway==TRUE) {
        for(j in 1:J) {
            eta[j,] <- apply(y[ (0:(M-1)*J) + j,  ], 2, mean) - mu
        }
    }

###    Calculate residuals by subtracting visit-specific mean from original functions for 
###    "twoway=TRUE", or subtracting overall mean function for "twoway=FALSE". 
    resd <- matrix(0, nrow=M*J, ncol=N) 
    resd <- t( t(y) - mu ) 
    if(twoway==TRUE) {
        for(j in 1:J) 
            resd[ (0:(M-1)*J) + j,  ] <- t ( t( y[ (0:(M-1)*J) + j,  ] ) - (mu + eta[j,]) ) 
    }

###     Calculate the pairwise difference between different visits within the same subject.
###     There are J(J-1)/2 possible pairs within the same subject.
    resd.pair.diff <- matrix(0, M*J*(J-1)/2, ncol=N)
    index <- matrix(0, J*(J-1)/2, 2)
    cur.index <- 1
    for(k in 1:(J-1)) 
        for(l in (k+1):J) {
            index[cur.index,] <- c(k,l)
            cur.index <- cur.index + 1
        }
    for(m in 1:M) {
        for(k in 1: (J*(J-1)/2)) {
            resd.pair.diff[ (m-1)*J*(J-1)/2 + k, ] <- resd[(m-1)*J + index[k,1], ] - resd[(m-1)*J + index[k,2], ]
        }
    }

###     Estimate the three covariance functions: overall covariance G, 
###     between covariance Gb and within covariance Gw
    G <- matrix(0, N, N)
    Gb <- matrix(0, N, N)
    Gw <- matrix(0, N, N)
    for(i in 1:N) 
        for(j in i:N) {
            G[i,j] <- cov(resd[,i],resd[,j])
            G[j,i] <- G[i,j]
            Gw[i,j] <- cov(resd.pair.diff[,i],resd.pair.diff[,j]) /2
            Gw[j,i] <- Gw[i,j]
        }
        
    Gb = G - Gw    

###     If "smoothing=TRUE" (which is recommended when functions are measured with noise), 
###     the program will smooth Gw(s,t) and Gb(s,t) before carrying out principal component
###     analysis. This step will reduce the bias in estimating eigenvalues and generate 
###     smooth eigenfunctions. 
    if(smoothing==TRUE) {

    ### drop the diagonal elements in G2 in smoothing step (the reason of doing so is
    ### explained in the paper)
        gw.temp <- Gw
        diag(gw.temp) <- rep(NA, N)
       
        data.gb <- data.frame(gb=as.vector(Gb),gw=as.vector(gw.temp),
                    x1=rep(seq(0,1,length=N), N), x2=rep(seq(0,1,length=N), each=N))
        attach(data.gb)
        
    ### load the "SemiPar" package which contains functions for semiparametric smoothing 
    ### using linear mixed models
        library(SemiPar)
        
        ####??      Define my own knots, but there seems to be a bug
             #myknots <- list( myknots=data.frame(x1=rep(seq(0,1,length=15),each=15), x2=rep(seq(0,1,length=15),15)) )
             #attach(myknots)
             myknots <- data.frame(x1=rep(seq(0,1,length=15),each=15), x2=rep(seq(0,1,length=15),15)) 
             fit1<- spm(gb ~ f(x1, x2,knots=myknots))
             fit<- spm(gw ~ f(x1, x2,knots=myknots),omit.missing=T)        
        
        pred1 <- predict(fit1,newdata)
        newdata <- data.frame(x1=x1,x2=x2)
        pred <- predict(fit,newdata)

    
    ### obtain the smoothed covariance functions Gw(s,t) and Gb(s,t)
        var.noise <- mean( diag(Gw) - diag(matrix(pred,100)) )
        s.gw <-matrix(pred,100) 
        Gw <- (s.gw + t(s.gw) )/2 
        s.gb <- matrix(pred1, 100)
        Gb <- (s.gb + t(s.gb))/2
    }



###     Estimate eigen values and eigen functions at two levels by calling the 
###     eigen function (in R "base" package) on discretized covariance matrices.
    e1 <- eigen(Gb)
    e2 <- eigen(Gw)

###    transform the eigenvalues of the covariance matrices into eigenvalues
###    of original functions
    fpca1.value <- e1$values * tlength / N
    fpca2.value <- e2$values * tlength / N 
###     Keep only non-negative eigenvalues
    fpca1.value <- ifelse(fpca1.value>=0, fpca1.value, 0)
    fpca2.value <- ifelse(fpca2.value>=0, fpca2.value, 0)
###     Calculate the percentage of variance that are explained by the components
    percent1 <- (fpca1.value)/sum(fpca1.value)
    percent2 <- (fpca2.value)/sum(fpca2.value)
    
###     Decide the number of components that are kept at level 1 and 2. The general
###     rule is to stop at the component where the cumulative percentage of variance 
###     explained is greater than 90% and the variance explained by any single component
###     after is less than 1/N. The number of components are also no less than the 
###     pre-determined minimum values min.K1 or min.K2.
    K1 <- max( which(cumsum(percent1) < 0.9 | percent1 > 1/N ) + 1, min.K1 )
    K2 <- max( which(cumsum(percent2) < 0.9 | percent2 > 1/N ) + 1, min.K2 )

###     estimate eigen vectors for discretized covariance matrices and
###     transform them into norm one eigenfunctions
    fpca1.vectors <- e1$vectors[, 1:K1]*sqrt(N/tlength)
    fpca2.vectors <- e2$vectors[, 1:K2]*sqrt(N/tlength)

###     The eigenfunctions are unique only up to a change of signs.
###     Select the signs of eigenfunctions so that the integration over the domain 
###     is non-negative
    for(i in 1:K1) {
        v2 <- fpca1.vectors[,i]
        tempsign <- sum(v2)
        fpca1.vectors[,i] <- ifelse(tempsign<0, -1,1) * v2
    }
    for(i in 1:K2) {
        v2 <- fpca2.vectors[,i]
        tempsign <- sum(v2)
        fpca2.vectors[,i] <- ifelse(tempsign<0, -1,1) * v2
    }
    
#############################################################################
###     The following part will estimate the principal component scores at 
###     both levels using the method described in the paper
#############################################################################

###     First, calculate the inner product (the cosine of the angles) between 
###     level 1 eigenfunctions and level 2 eigenfunctions
    cross.integral <- matrix(0, K1, K2)
    for(i in 1:K1)
        for(j in 1:K2) 
            cross.integral[i,j] <- sum(fpca1.vectors[,i]* fpca2.vectors[,j]) *tlength/N
 
###     Next, calculate the inner product of each centered function with the 
###     level 1 or level 2 eigenfunctions
    int1 <- matrix(0, M*J, K1)
    int2 <- matrix(0, M*J, K2)
    for(i in 1:(M*J))   {
        for(j in 1:K1)  {
            int1[ i ,j] <- sum( resd[i,] * fpca1.vectors[,j] ) * tlength /N
        }
        for(j in 1:K2) {
            int2[ i ,j] <- sum( resd[i,] * fpca2.vectors[,j] ) * tlength /N    
        }
    }

    
###     Finally, calculate the principal component scores based on the formulas
###     given in the paper.
        s1 <- matrix(0, M*J, K1)
        s2 <- matrix(0, M*J, K2)
        library(MASS)
        design.xi <- ginv( diag(rep(1,K1)) - cross.integral %*% t(cross.integral) )
        for(m in 1:M) {
            resid <- rep(0, K1)
            for(j in 1:J) {
                index <- (m-1) * J + j
                resid <- resid + ( int1[index,] - drop(cross.integral %*% int2[index,]) )/J
            }
            index.m <- ( (m-1) * J + 1 ) : (m*J)
            xi.temp <- design.xi %*% resid
            s1[index.m,] <- matrix(rep(xi.temp, each=J), nrow=J)
            for(j in 1:J) {
                index <- (m-1) * J + j
                s2[index,] <- int2[index,] - drop( t(cross.integral) %*% xi.temp )
            }
        }


###     Return the results from multilevel FPCA as a list
    return( list(K1=K1, K2=K2, lambda1=fpca1.value, lambda2=fpca2.value, phi1=fpca1.vectors, 
        phi2=fpca2.vectors, scores1=s1, scores2=s2, mu=mu, eta=t(eta) ) )

} 
