

############################################################################
###
###     This file contains R codes to estimate the principal component scores
###         for Multilevel Functional Princiapl Component Analysis (MFPCA). 
###
###     The true data are generated from Case 2 in the paper, and the 
###     function "MFPCA" are called to fit the MFPCA model. The level 1 
###     eigenfunctions are four sine and cosine functions, and the level 2
###     eigenfunctions are muturally orthogonal polynomials.
###
###
###         Author: Chongzhi Di
###         Based on the work with Ciprian M.  Crainiceanu 
###                                                             
###         The corresponding paper is under review
###         Date:       05/05/2008
##########################################################################



###     Set the seed for simulation
set.seed(1371)

###     K1: number of true principal components
###     K2: number of true principal components 
###     K = K1 + K2 : number of total principal components
###     The true data is generated with K1=4 and K2=4

###     N: number of grid points for each function
###     M: number of clusters of functions, or number of subjects
###     J: number of functions per cluster, or number of visits per subject

###     t: actual knots for each function
###     lambda: true eigen values
###     f:      true eigen functions
###     sigma:  standard deviation of measurement error

K1 <- 4
K2 <- 4
K <- K1 + K2
N <- 100
sigma <- 1

###     generate the true eigenvalues
lambda1 <-  0.5^(0:(K1-1))
lambda2 <-  0.5^(0:(K2-1))

###     generate the set of equally spaced grid points on which 
###     the functions are measured
tlength <- 1
t <- seq(0,tlength,length=N)


###     generate bases of the functional space
###     K base functions

###     The true level 1 eigenfunctions are four muturally orthogonal 
###     sine and cosine functions
f1 <- matrix(0,nrow=K1,ncol=N)
for ( i in 1:(K1/2) ) {
    f1[2*i-1,] <- sqrt(2/tlength)*sin(i*t*2*pi/tlength)
    f1[2*i,] <- sqrt(2/tlength)*cos(i*t*2*pi*tlength)
}


###     The true level 2 eigenfunctions are four muturally orthogonal 
###     polynomial functions
tt <- t/tlength
f2 <- matrix(0, nrow=K2, ncol=N)
f2[1,] <- rep(1, N)*sqrt(1/tlength)
f2[2,] <- sqrt(3/tlength) * (2*tt - 1)
f2[3,] <- sqrt(5/tlength) * (6*tt^2 - 6 * tt + 1)
f2[4,] <- sqrt(7/tlength) * (20*tt^3 - 30*tt^2 + 12 * tt -1)

f <- rbind(f1, f2)


###     Calculate the paiwise inner products between eigenfunctions
###     By construction, the norm of each function is 1, i.e, the inner product
###     between each eigenfunction and itself is 1. 
###
###     The inner product between any two functions is the cosine of the angle 
###     between them. The farther away the inner product is from 0, the more they
###     are departed from orthogonal. 
norm <- matrix(0, K, K)
for(i in 1:K)
    for(j in 1:K)
        norm[i,j] <- sum( f[i,] * f[j,] )*tlength /N




###     Generate M random coefficients si's from normal distributions 
###     with variance equal to the corresponding eigenvalue.
M <- 200
J <- 3
si1 <- matrix(0, nrow=M, ncol=K1)
si2 <- matrix(0, nrow=M*J, ncol=K2)
for(k in 1:K1) {
    si1[,k] <- rnorm( M,sd=sqrt(lambda1[k]) )
}
for(k in 1:K2) {
    si2[,k] <- rnorm( M * J ,sd=sqrt(lambda2[k]) )
}

###     Calculate the random functions as the sum of the products of generated scores
###     and corresponding eigenfunctions. 
y <- matrix(0,nrow=M*J,ncol=N)
for(m in 1:M) {
    temp <- apply( ( si1[m,] %*% t(rep(1,N)) ) * f1, 2,sum)
    for(j in 1:J) {
        y[ (m-1)*J + j ,] <- temp + apply( ( si2[(m-1)*J + j ,] %*% t(rep(1,N)) ) * f2, 2,sum) + rnorm(N, sd=sigma)
    }
}








###########################################################################################
####    The following codes would perform Multilevel FPCA on both between and within levels
####
####    Alternatively, one could also call the function "MFPCA" as follows
####    
####        setwd("C:/Documents and Settings/Cdi/My Documents/_Sleep/sleepsoftware/FPCA-paper-doc/")
####        source("MFPCA.R")
####        fit <- MFPCA(y, M=M, J=J, N=N, tlength=1) 
####    
###########################################################################################

mu <- apply(y, 2,mean)
eta <- matrix(0, J, N)
for(j in 1:J) {
    eta[j,] <- apply(y[ (0:(M-1)*J) + j,  ], 2, mean) - mu
}

###    Calculate residuals by subtracting visit-specific mean from original functions 
resd <- matrix(0, nrow=M*J, ncol=N) 
for(j in 1:J) 
    resd[ (0:(M-1)*J) + j,  ] <- t ( t( y[ (0:(M-1)*J) + j,  ] ) - (mu + eta[j,]) ) 
resd.rev <- matrix(0, nrow=M*J, ncol=N) 
for(m in 1:M) {
    resd.rev[ (m-1)*J + 1:2 ,] <- resd[(m-1)*J + 2:1, ]
}

###     Calculate the pairwise difference between different visits within the same subject.
###     There are J(J-1)/2 possible pairs within the same subject.
    resd.pair.diff <- matrix(0, M*J*(J-1)/2, ncol=N)
    index <- matrix(0, J*(J-1)/2, 2)
    cur.index <- 1
    for(k in 1:(J-1)) 
        for(l in (k+1):J) {
            index[cur.index,] <- c(k,l)
            cur.index <- cur.index + 1
        }
    for(m in 1:M) {
        for(k in 1: (J*(J-1)/2)) {
            resd.pair.diff[ (m-1)*J*(J-1)/2 + k, ] <- resd[(m-1)*J + index[k,1], ] - resd[(m-1)*J + index[k,2], ]
        }
    }


###     Estimate the three covariance functions: overall covariance G, 
###     between covariance Gb and within covariance Gw
    G <- matrix(0, N, N)
    G1 <- matrix(0, N, N)
    G2 <- matrix(0, N, N)
    for(i in 1:N) 
        for(j in i:N) {
            G[i,j] <- cov(resd[,i],resd[,j])
            G[j,i] <- G[i,j]
            G1[i,j] <- cov(resd.pair.diff[,i],resd.pair.diff[,j]) /2
            G1[j,i] <- G1[i,j]
        }
        
    G2 = G - G1

##### smooth the kernels in the presence of noise
    g2.temp <- G2
    diag(g2.temp) <- rep(NA, N)
    g2 <- as.vector(g2.temp)
    g1 <- as.vector(G1)
    x1 <- rep(seq(0,1,length=N), N)
    x2 <- rep(seq(0,1,length=N), each=N)
    library(SemiPar)

    myknots <- data.frame(x1=rep(seq(0,1,length=15),each=15), x2=rep(seq(0,1,length=15),15))
    fit<- spm(g2 ~ f(x1, x2,knots=myknots),omit.missing=T)
    newdata <- data.frame(x1=x1,x2=x2)
    pred <- predict(fit,newdata)
    fit1<- spm(g1 ~ f(x1, x2,knots=myknots))
    pred1 <- predict(fit1,newdata)


    s2 <- mean( diag(G2) - diag(matrix(pred,100)) )
    s.g2 <-matrix(pred,100) 
    G2 <- (s.g2 + t(s.g2) )/2 
    s.g1 <- matrix(pred1, 100)
    G1 <- (s.g1 + t(s.g1))/2

####    estimate eigen values and eigen funcitons at two levels
fpca1.value <- eigen(G1)$values[1:(K1)]/N
fpca2.value <- eigen(G2)$values[1:(K2)]/N

rbind(fpca1.value, lambda1, fpca2.value, lambda2)

fpca1.vectors <- eigen(G1)$vectors[, 1:K1]*sqrt(N)
fpca2.vectors <- eigen(G2)$vectors[, 1:K2]*sqrt(N)

for(i in 1:K1) {
    v1 <- f1[i,]
    v2 <- fpca1.vectors[,i]
    tempsign <- sum((v1-v2)^2) - sum((v1+v2)^2)
    fpca1.vectors[,i] <- ifelse(tempsign>0, -1,1) * v2
}
for(i in 1:K2) {
    v1 <- f2[i,]
    v2 <- fpca2.vectors[,i]
    tempsign <- sum((v1-v2)^2) - sum((v1+v2)^2)
    fpca2.vectors[,i] <- ifelse(tempsign>0, -1,1) * v2
}




###########################################################################################
####    Estimate the principal component scores by Bayesian MCMC
####    Method 1: the full model (PC-F in the paper)      
####
####        Eigenvalues and eigenfunctions are fixed
####        Measurement error variance would be estimated from the model
###########################################################################################


####    Call the "R2WinBUGS" package
library(R2WinBUGS)

####    Set eigenvalues and eigenfunctions as estimated previousely
phi1 <- t(fpca1.vectors)
phi2 <- t(fpca2.vectors)
tau1 <- 1/fpca1.value
tau2 <- 1/fpca2.value


####    Set up the data set and initial values for MCMC
data <- list("M","J","N","K1","K2","y","phi1", "phi2", "tau1", "tau2")
inits <- function(){
    list(tau=1 )
}  

####    Call "bugs" functions to run MCMC via WinBUGS
####    We would use 2500 iterations with 500 burn-ins
setwd("C:\\Documents and Settings\\Cdi\\My Documents\\_Sleep\\WinBUGS\\")
sim.m1 <- bugs(data, inits=inits, model.file = "scores-full.odc",
    parameters = c("sigma", "s1","s2"),
    n.chains = 1, n.iter = 2500, n.burnin=500, n.thin=1,
    bugs.directory = "c:/Program Files/WinBUGS14/",debug=F)
    
    
####    Export and summarize results from MCMC
####        "summ.m1" would contain the posterior mean, SD, "2.5%", "50%", and "97.5%" pencentiles
####        for every parameter (noise variance and PC scores)
summ.m1 <- matrix(0, M*(K1+K2*J)+1, 5)
rownames(summ.m1) <- colnames(sim.m1$sims.matrix)[1:(M*(K1+K2*J)+1)]
colnames(summ.m1) <- c("mean", "sd", "2.5%", "50%", "97.5%")
summ.m1[,1] <- apply(sim.m1$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, mean)
summ.m1[,2] <- apply(sim.m1$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, sd)
summ.m1[,3] <- apply(sim.m1$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, quantile, prob=0.025)
summ.m1[,4] <- apply(sim.m1$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, quantile, prob=0.50)
summ.m1[,5] <- apply(sim.m1$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, quantile, prob=0.975)
round(summ.m1, digits=3)[1:10,]

####    Transform the posterior mean and SD into matrix form
####    "si1.est1" contains the posterior mean for level 1 PC scores for every subject. 
####        Each row corresponds to a subject, and each column corresponds to a component. 
####    "si2.est1" contains the posterior mean for level 2 PC scores for every subject and visit. 
####    "si1.sd.est1" contains the posterior SD for level 1 PC scores for every subject. 
####    "si2.sd.est1" contains the posterior SD for level 2 PC scores for every subject and visit. 
si1.est1 <- matrix(summ.m1[2:(M*K1+1),1], nrow=M, K1, byrow=T)
si2.est1 <- matrix(summ.m1[(M*K1+2):(M*K1+M*K2*J+1),1], nrow=M*J, K2, byrow=T)
si1.sd.est1 <- matrix(summ.m1[2:(M*K1+1),2], nrow=M, K1, byrow=T)
si2.sd.est1 <- matrix(summ.m1[(M*K1+2):(M*K1+M*K2*J+1),2], nrow=M*J, K2, byrow=T)


####    plot the Markov chain obtained for parameter "sigma"
####    This would correspond to Figure 4 in the supplementary file  
plot(sim.m1$sims.matrix[,1], ylim=c(0.98, 1.02), main="sigma", type="l")
lines(c(-1, 1000), c(1,1), lty=2, col="red")

####    plot the Markov chain obtained for level 1 principal component scores
####    This would correspond to Figure 5 in the supplementary file  
sd.temp <- rep(sqrt(fpca1.value), 3)
par(mfcol=c(4,2), mar=c(2,2,4,2))
for(i in 2:9) {
plot(sim.m1$sims.matrix[,i], ylim=c(summ.m1[i,1]-2*sd.temp[i-1], summ.m1[i,1]+2*sd.temp[i-1]), 
    main=colnames(sim.m1$sims.matrix)[i], type="l")
lines(c(-1, 2001), rep(as.vector(t(si1))[i-1],2),lty=2,col="red")
}

####    plot the Markov chain obtained for level 2 principal component scores
####    This would correspond to Figure 6 in the supplementary file  
sd.temp <- rep(sqrt(fpca2.value), 3)
par(mfcol=c(4,2), mar=c(2,2,4,2))
for(i in 802:809) {
plot(sim.m1$sims.matrix[,i], ylim=c(summ.m1[i,1]-2*sd.temp[i-801], summ.m1[i,1]+2*sd.temp[i-801]), 
    main=colnames(sim.m1$sims.matrix)[i], type="l")
lines(c(-1, 2001), rep(as.vector(t(si2))[i-801],2),lty=2, col="red")
}



###############################################################################################
###     The previous parts use a full Bayesian approach to estimate the PC scores
###         by BLP (best linear prediction). The approach is valid for both dense
###         and sparse grids cases, but computational intensive for dense grids.
###                                                                             
###############################################################################################



###########################################################################################
####    Estimate the principal component scores by Bayesian MCMC
####    Method 2: the projection model (PC-P in the paper)      
####
####        Eigenvalues and eigenfunctions are fixed
####        Measurement error variance would be estimated from the model
###########################################################################################


####    First, project each function (per subject per visit) onto the functional space spanned 
####        by the level 1 eigenfunctions and the space spanned by the level 2 eigenfunctions
mc <- matrix(0, K1, K2)
a <- matrix(0, M*J, K1)
b <- matrix(0, M*J, K2) 

####    calculate the inner product between level 1 and level 2 eigenfunctions
    for(i in 1:K1)
        for(j in 1:K2) 
            mc[i,j] <- sum(fpca1.vectors[,i]* fpca2.vectors[,j]) /N
 
####    calculate the projections of each function onto the level 1 and level 2 eigen functional space
    for(i in 1:(M*J))   {
        for(j in 1:K1)  {
            a[ i ,j] <- sum( resd[i,] * fpca1.vectors[,j] ) /N
        }
        for(j in 1:K2) {
            b[ i ,j] <- sum( resd[i,] * fpca2.vectors[,j] ) /N    
        }
    }

####    Set eigenvalues and eigenfunctions as estimated previousely
phi1 <- t(fpca1.vectors)
phi2 <- t(fpca2.vectors)
tau1 <- 1/fpca1.value
tau2 <- 1/fpca2.value


####    Set up the data set and initial values for MCMC
data3 <- list("M","J","K1","K2","a","b","mc", "tau1", "tau2")
inits3 <- function(){
    list(tau=1 )
}  

####    Call "bugs" functions to run MCMC via WinBUGS
####    We would use 2500 iterations with 500 burn-ins
sim.m3 <- bugs(data3, inits=inits3, model.file = "scores-blp.odc",
    parameters = c("sigma", "s1","s2"),
    n.chains = 1, n.iter = 2500, n.burnin=500, n.thin=1,
    bugs.directory = "c:/Program Files/WinBUGS14/",debug=F)


 
####    Export and summarize results from MCMC
####        "summ.m3" would contain the posterior mean, SD, "2.5%", "50%", and "97.5%" pencentiles
####        for every parameter (noise variance and PC scores)
summ.m3 <- matrix(0, M*(K1+K2*J)+1, 5)
rownames(summ.m3) <- colnames(sim.m3$sims.matrix)[1:(M*(K1+K2*J)+1)]
colnames(summ.m3) <- c("mean", "sd", "2.5%", "50%", "97.5%")
summ.m3[,1] <- apply(sim.m3$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, mean)
summ.m3[,2] <- apply(sim.m3$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, sd)
summ.m3[,3] <- apply(sim.m3$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, quantile, prob=0.025)
summ.m3[,4] <- apply(sim.m3$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, quantile, prob=0.50)
summ.m3[,5] <- apply(sim.m3$sims.matrix[,1:(M*(K1+K2*J)+1)], 2, quantile, prob=0.975)
round(summ.m3, digits=3)[1:10,]

####    Transform the posterior mean and SD into matrix form
####    "si1.est3" contains the posterior mean for level 1 PC scores for every subject. 
####        Each row corresponds to a subject, and each column corresponds to a component. 
####    "si2.est3" contains the posterior mean for level 2 PC scores for every subject and visit. 
####    "si1.sd.est3" contains the posterior SD for level 1 PC scores for every subject. 
####    "si2.sd.est3" contains the posterior SD for level 2 PC scores for every subject and visit. 
si1.est3 <- matrix(summ.m3[2:(M*K1+1),1], nrow=M, K1, byrow=T)
si2.est3 <- matrix(summ.m3[(M*K1+2):(M*K1+M*K2*J+1),1], nrow=M*J, K2, byrow=T)
si1.sd.est3 <- matrix(summ.m3[2:(M*K1+1),2], nrow=M, K1, byrow=T)
si2.sd.est3 <- matrix(summ.m3[(M*K1+2):(M*K1+M*K2*J+1),2], nrow=M*J, K2, byrow=T)


####    plot the Markov chain obtained for parameter "sigma"
plot(sim.m3$sims.matrix[,1], ylim=c(0.045, 0.051), main="sigma", type="l")

####    plot the Markov chain obtained for level 1 principal component scores
sd.temp <- rep(sqrt(fpca1.value), 3)
par(mfcol=c(4,2), mar=c(2,2,4,2))
for(i in 2:9) {
plot(sim.m3$sims.matrix[,i], ylim=c(summ.m3[i,1]-2*sd.temp[i-1], summ.m3[i,1]+2*sd.temp[i-1]), 
    main=colnames(sim.m3$sims.matrix)[i], type="l")
lines(c(-1, 2001), rep(as.vector(t(si1))[i-1],2),lty=2,col="red")
}

####    plot the Markov chain obtained for level 2 principal component scores
sd.temp <- rep(sqrt(fpca2.value), 3)
par(mfcol=c(4,2), mar=c(2,2,4,2))
for(i in 802:809) {
plot(sim.m3$sims.matrix[,i], ylim=c(summ.m3[i,1]-2*sd.temp[i-801], summ.m3[i,1]+2*sd.temp[i-801]), 
    main=colnames(sim.m3$sims.matrix)[i], type="l")
lines(c(-1, 2001), rep(as.vector(t(si2))[i-801],2),lty=2, col="red")
}





####################        Compare the MSE of the two methods in estimating scores
res <- matrix(0, 8, 4)
colnames(res) <- c("comp.1", "comp.2", "comp.3", "comp.4")
rownames(res) <- rep( c("rmse.m1", "rmse.m2", "rmse.m3","rmse.m4"), 2) 

    temp11 <- si1.est1 - si1
    temp31 <- si1.est3 - si1   
    res[1,] <- sqrt( apply(temp11^2, 2, mean) )
    res[2,] <- sqrt( apply(temp31^2, 2, mean) )

    temp12 <- si2.est1 - si2
    temp22 <- si2.est3 - si2
   
    res[3,] <- sqrt( apply(temp12^2, 2, mean) )
    res[4,] <- sqrt( apply(temp32^2, 2, mean) )
