
############################################################################
###
###     This file contains R codes to conduct simulation studies to 
###     evaluate the performance of the MFPCA. 
###
###     The true data are generated from Case 2 in the paper, and the 
###     function "MFPCA" are called to fit the MFPCA model. The level 1 
###     eigenfunctions are four sine and cosine functions, and the level 2
###     eigenfunctions are muturally orthogonal polynomials.
###
###
###         Author: Chongzhi Di
###         Based on the work with Ciprian M.  Crainiceanu 
###                                                             
###         The corresponding paper is under review
###         Date:       06/20/2007
##########################################################################


############## simulation of multilevel functional data

###     Set the seed for simulation
set.seed(1371)

###     K1: number of true principal components
###     K2: number of true principal components 
###     K = K1 + K2 : number of total principal components
###     The true data is generated with K1=4 and K2=4

###     N: number of grid points for each function
###     M: number of clusters of functions
###     J: number of functions per cluster

###     t: actual knots for each function
###     lambda: true eigen values
###     f:      true eigen functions

K1 <- 4
K2 <- 4
K <- K1 + K2
N <- 100

###     generate the true eigenvalues
lambda1 <-  0.5^(0:(K1-1))
lambda2 <-  0.5^(0:(K2-1))

###     generate the set of equally spaced grid points on which 
###     the functions are measured
tlength <- 1
t <- seq(0,tlength,length=N)


###     generate bases of the functional space
###     K base functions

###     The true level 1 eigenfunctions are four muturally orthogonal 
###     sine and cosine functions
f1 <- matrix(0,nrow=K1,ncol=N)
for ( i in 1:(K1/2) ) {
    f1[2*i-1,] <- sqrt(2/tlength)*sin(i*t*2*pi/tlength)
    f1[2*i,] <- sqrt(2/tlength)*cos(i*t*2*pi*tlength)
}


###     The true level 2 eigenfunctions are four muturally orthogonal 
###     polynomial functions
tt <- t/tlength
f2 <- matrix(0, nrow=K2, ncol=N)
f2[1,] <- rep(1, N)*sqrt(1/tlength)
f2[2,] <- sqrt(3/tlength) * (2*tt - 1)
f2[3,] <- sqrt(5/tlength) * (6*tt^2 - 6 * tt + 1)
f2[4,] <- sqrt(7/tlength) * (20*tt^3 - 30*tt^2 + 12 * tt -1)

f <- rbind(f1, f2)


###     Calculate the paiwise inner products between eigenfunctions
###     By construction, the norm of each function is 1, i.e, the inner product
###     between each eigenfunction and itself is 1. 
###
###     The inner product between any two functions is the cosine of the angle 
###     between them. The farther away the inner product is from 0, the more they
###     are departed from orthogonal. 
norm <- matrix(0, K, K)
for(i in 1:K)
    for(j in 1:K)
        norm[i,j] <- sum( f[i,] * f[j,] )*tlength /N


###     Generate M random coefficients si's from normal distributions 
###     with variance equal to the corresponding eigenvalue.
M <- 200
J <- 3
si1 <- matrix(0, nrow=M, ncol=K1)
si2 <- matrix(0, nrow=M*J, ncol=K2)
for(k in 1:K1) {
    si1[,k] <- rnorm( M,sd=sqrt(lambda1[k]) )
}
for(k in 1:K2) {
    si2[,k] <- rnorm( M * J ,sd=sqrt(lambda2[k]) )
}

###     Calculate the random functions as the sum of the products of generated scores
###     and corresponding eigenfunctions. 
y <- matrix(0,nrow=M*J,ncol=N)
for(m in 1:M) {
    temp <- apply( ( si1[m,] %*% t(rep(1,N)) ) * f1, 2,sum)
    for(j in 1:J) {
        y[ (m-1)*J + j ,] <- temp + apply( ( si2[(m-1)*J + j ,] %*% t(rep(1,N)) ) * f2, 2,sum)
    }
}


###     Plot generated functions in the first 3 clusters
par(mfrow=c(3,2))
for(i in 1:6) {
    plot(t,t,type="n",ylim=c(-6,6),ylab="",main=paste("subject",i))
    for(j in 1:J) {
        lines(t,y[(i-1) * J + j,],col=ifelse(j==2,"red","black"))
    }
}
par(mfrow=c(1,1))




###     Call the "MFPCA" function to carry out multilevel FPCA
###     Change the working directory to the one that contains the "MFPCA.R" file
setwd("C:/Documents and Settings/Cdi/My Documents/_Sleep/sleepsoftware/FPCA-paper-doc/")
source("MFPCA.R")

fit <- MFPCA(y, M=M, J=J, N=N, tlength=1) 

# fit2 <- MFPCA(y, M=M, J=J, N=N, tlength=1, smoothing=T) 



###     Plot the estimated mean functions
plot(t, fit$mu, xlab="t", ylab="Mean functions", ylim=c(-6,6), type="l",lwd=2 )
for(j in 1:J) {
    lines(t, fit$eta[,j], lty=j+1)
}
legend(0,-2,bty="n",legend=c("Overall",paste("visit",1:J)), lty=1:(J+1), lwd=c(2,rep(1,J)))

###     Output the level 1 and 2 eigenvalues and compare with true eigenvalues
fit.lambda1 <- fit$lambda1[1:fit$K1]
fit.lambda2 <- fit$lambda2[1:fit$K2]

lambda1 <- data.frame(estimate=fit.lambda1[1:K1], true=lambda1)
lambda2 <- data.frame(estimate=fit.lambda2[1:K2], true=lambda2)
lambda1
lambda2

###     Plot the estimated level 1 eigenfunctions vs. true eigenfunctions
###     Choose the sign that makes the two as close as possible, since the 
###     eigenfunctions are unique only up to the change of signs.
par(mfrow=c(2,2))
for(i in 1:4) {
    v1 <- f1[i,]
    v2 <- fit$phi1[,i]
    tempsign <- sum((v1-v2)^2) - sum((v1+v2)^2)
    v2 <- ifelse(tempsign>0, -1,1) * v2
    plot(t,v2, ylim=c(-1.5,1.5), main="level 1 eigenfunctions", ylab="",xlab="t")
    lines(t,v1, col="red", lwd=2)
}
par(mfrow=c(1,1))

###     Plot the estimated level 2 eigenfunctions vs. true eigenfunctions
###     Choose the sign that makes the two as close as possible, since the 
###     eigenfunctions are unique only up to the change of signs.
par(mfrow=c(2,2))
for(i in 1:4) {
    v1 <- f2[i,]
    v2 <- fit$phi2[,i]
    tempsign <- sum((v1-v2)^2) - sum((v1+v2)^2)
    v2 <- ifelse(tempsign>0, -1,1) * v2
    plot(t,v2, ylim=c(-1.5,1.5), main="level 1 eigenfunctions", ylab="",xlab="t")
    lines(t,v1, col="red", lwd=2)
}
par(mfrow=c(1,1))
